#include "bs.h"
#include <gdk/gdkkeysyms.h>
#include <ctype.h>



/* Draws form's field (with text) in backing store */
static void field_display( Form *form, int field_num )
{
    FormField *field = NULL;

    assert(form);
    assert(field_num >= 0 && field_num < NUM_FIELDS);
    assert(form->fields + field_num);
    field = form->fields + field_num;

    /* Blank out the field's rectangle */
    gdk_draw_rectangle( form->pixmap, form->bg_gc, TRUE,
                        form->position.x + field->position.x - FIELD_X_PAD,
                        form->position.y + field->position.y - FIELD_Y_PAD,
                        field->width * font_char_width + 2 * FIELD_Y_PAD,
                        font_char_height + 2 * FIELD_Y_PAD );
    /* Draw the field's bounding box */
    gdk_draw_rectangle( form->pixmap, form->fg_gc, FALSE,
                        form->position.x + field->position.x - FIELD_X_PAD,
                        form->position.y + field->position.y - FIELD_Y_PAD,
                        field->width * font_char_width + 2 * FIELD_Y_PAD,
                        font_char_height + 2 * FIELD_Y_PAD );
    /* Draw the field's name */
    gdk_draw_string( form->pixmap, font, form->fg_gc,
                     form->position.x + field->position.x,
                     form->position.y + field->position.y + font_char_height,
                     form->curr_tuple->fields[field_num] );
    /* Draw the field's text */
    if ( form->curr_tuple )
        gdk_draw_string( form->pixmap, font, form->fg_gc,
                         FORM_X_PAD,
                         form->position.y + field->position.y + font_char_height,
                         field_dsc[field_num].name );

    return;
}


/* Repaints current field right of and including cursor position, both in
 * backing store and on the screen.  Does not hide or redraw cursor--do it
 * yourself */
static void form_update_curr_field( Form *form )
{
    FormField *field;
    Tuple     *tuple;
    int        len;

#define FIELD_BORDER_THICKNESS      1

    assert(form);
    assert(form->fields + form->curr_field);
    assert(form->curr_tuple);
    field = form->fields + form->curr_field;
    tuple = form->curr_tuple;
    len   = strlen( *(tuple->fields + form->curr_field) );

    /* Blank out backing store from the current cursor position to the end of
     * the field */
    gdk_draw_rectangle( form->pixmap, form->bg_gc, TRUE,
                        form->position.x + field->position.x + form->cursor * font_char_width,
                        form->position.y + field->position.y - FIELD_Y_PAD + FIELD_BORDER_THICKNESS,
                        (field->width - form->cursor) * font_char_width,
                        font_char_height + 2 * FIELD_Y_PAD - 2*FIELD_BORDER_THICKNESS + 1);
    /* Draw text in the backing store */
    gdk_draw_string( form->pixmap, font, form->fg_gc,
                     form->position.x + field->position.x + form->cursor * font_char_width,
                     form->position.y + field->position.y + font_char_height,
                     form->curr_tuple->fields[form->curr_field] + form->cursor );
    /* Blank out screen from the current cursor position to the end of the
     * field */
    gdk_draw_rectangle( form->screen, form->bg_gc, TRUE,
                        form->position.x + field->position.x + form->cursor * font_char_width,
                        form->position.y + field->position.y - FIELD_Y_PAD + FIELD_BORDER_THICKNESS,
                        (field->width - form->cursor) * font_char_width,
                        font_char_height + 2 * FIELD_Y_PAD - 2*FIELD_BORDER_THICKNESS + 1);
    /* Draw text on the screen */
    gdk_draw_string( form->screen, font, form->fg_gc,
                     form->position.x + field->position.x + form->cursor * font_char_width,
                     form->position.y + field->position.y + font_char_height,
                     form->curr_tuple->fields[form->curr_field] + form->cursor );

    return;
}


/* Draws form and each of its fields in backing store */
static void form_display( Form *form )
{
    int i;
    gint width, height;

    assert(form);
    assert(form->pixmap);

    gdk_window_get_size( form->pixmap, &width, &height );

    /* Blank out backing store */
    gdk_draw_rectangle( form->pixmap, form->bg_gc, TRUE, 0, 0, width, height );

    for ( i = 0; i < NUM_FIELDS; i++ )
        field_display( form, i );

    return;
}


/* Inverts image of cursor in the current cursor position of the form */
static void form_draw_cursor( Form *form )
{
    gint x, y;

    assert(form);

    if ( form->cursor >= 0 ) {
        gdk_gc_set_function( form->fg_gc, GDK_INVERT );
        x = form->position.x + form->fields[form->curr_field].position.x
            + form->cursor * font_char_width;
        y = form->position.y + form->fields[form->curr_field].position.y
            - CURSOR_Y_PAD + 1;
        /* Draw cursor in backing store */
        gdk_draw_rectangle( form->pixmap, form->fg_gc, TRUE, x, y,
                            font_char_width,
                            font_char_height + 2 * CURSOR_Y_PAD - 1 );
        /* Draw cursor on screen */
        gdk_draw_rectangle( form->screen, form->fg_gc, TRUE, x, y,
                            font_char_width,
                            font_char_height + 2 * CURSOR_Y_PAD - 1);
        gdk_gc_set_function( form->fg_gc, GDK_COPY );
    }

    return;
}


void form_process_char( Form *form, guint state, guint keyval, char character )
{
    FormField *field;
    char      *string;
    int        max_width;

    assert(form);
    field = &form->fields[form->curr_field];

    if ( form->cursor < 0 ) {
        form->cursor = 0;
        form_draw_cursor(form);
        return;
    }

    string = form->curr_tuple->fields[form->curr_field];
    max_width = field->width; /* should be the same both for form's field width
                               * and maximum width of text for the field */

    /*
     * Control key is down
     */
    if ( !(state & GDK_SHIFT_MASK) &&
         (state & GDK_CONTROL_MASK)  &&
         !(state & GDK_MOD1_MASK) ) {
        switch ( keyval ) {
        case GDK_F: case GDK_f:
            goto MoveRight;
            break;
        case GDK_B: case GDK_b:
            goto MoveLeft;
            break;
        case GDK_E: case GDK_e:
            goto MoveEnd;
            break;
        case GDK_A: case GDK_a:
            goto MoveHome;
            break;
        case GDK_N: case GDK_n:
            goto NextField;
            break;
        case GDK_P: case GDK_p:
            goto PrevField;
            break;
        case GDK_D: case GDK_d:
            goto DelChar;
            break;
        case GDK_K: case GDK_k:
            goto KillToEnd;
            break;
        case GDK_U: case GDK_u:
            goto KillToStart;
            break;
        default:
            return;
        }
    }

    /*
     * Modifier key is down
     */
    if ( !(state & GDK_SHIFT_MASK) &&
         !(state & GDK_CONTROL_MASK) &&
         (state & GDK_MOD1_MASK) ) {
        switch ( keyval ) {
        case GDK_F: case GDK_f:
            goto MoveWordRight;
            break;
        case GDK_B: case GDK_b:
            goto MoveWordLeft;
            break;
        case GDK_D: case GDK_d:
            goto DelWord;
            break;
        case GDK_BackSpace:
            goto BSWord;
            break;
        default:
            return;
        }
    }

    /*
     * Shift key is down
     */
    if ( (state & GDK_SHIFT_MASK) &&
         !(state & GDK_CONTROL_MASK) &&
         !(state & GDK_MOD1_MASK) ) {
        switch ( keyval ) {
        case GDK_Tab: case GDK_ISO_Left_Tab:
            goto PrevField;
            break;
/*         default: */
/*             return; */
        }
    }

    /*
     * Control and modifier keys are up
     */
    if ( !(state & GDK_CONTROL_MASK) &&
         !(state & GDK_MOD1_MASK) ) {
        switch ( keyval ) {
        case GDK_Right:
            goto MoveRight;
            break;
        case GDK_Left:
            goto MoveLeft;
            break;
        case GDK_End:
            goto MoveEnd;
            break;
        case GDK_Home:
            goto MoveHome;
            break;
        case GDK_Tab: case GDK_Down:
            goto NextField;
            break;
        case GDK_Up:
            goto PrevField;
            break;
        case GDK_Delete:
            goto DelChar;
            break;
        case GDK_BackSpace:
            goto BSChar;
            break;
        default:
            /* If the fields has no filter, accept the character if it is a
             * graphical character or a space; if the field has a filter, accept
             * the character if it gets filtered thru */
            if ( ( field_dsc[form->curr_field].filter == NULL &&
                   (isgraph(character) || character == ' ')       )   ||
                 ( field_dsc[form->curr_field].filter != NULL &&
                   (*field_dsc[form->curr_field].filter)(character) )    ) {
                int width = strlen(string);
                int curr;

                if ( width >= max_width )
                    return;

                /* Move all chars (from trailing '\0' to the current) one char
                 * space to the right */
                curr = width;
                do
                    *(string + width + 1) = *(string + width);
                while ( width-- > form->cursor );
                /* Write new char */
                *(string + form->cursor) = character;
                form_draw_cursor(form);
                form_update_curr_field( form );
                if ( max_width > form->cursor + 1 )
                    form->cursor++;
                form_draw_cursor(form);
            }
            return;
        }
    }

    return;

  MoveRight:
    if ( max_width > form->cursor + 1 &&
         *(string + form->cursor) != '\0' ) {
        form_draw_cursor(form);
        form->cursor += 1;
        form_draw_cursor(form);
    }
    return;

  MoveLeft:
    if ( form->cursor > 0 ) {
        form_draw_cursor(form);
        form->cursor -= 1;
        form_draw_cursor(form);
    }
    return;

  MoveEnd:
    form_draw_cursor(form);
    while ( max_width > form->cursor + 1 &&
            *(string + form->cursor) != '\0' )
        form->cursor++;
    form_draw_cursor(form);
    return;

  MoveHome:
    form_draw_cursor(form);
    form->cursor = 0;
    form_draw_cursor(form);
    return;

  MoveWordRight:
    form_draw_cursor(form);
    /* Skip all blanks */
    while ( max_width > form->cursor + 1 &&
            *(string + form->cursor) != '\0' &&
            isspace(*(string + form->cursor)) )
        form->cursor++;
    /* Skip all non-blanks */
    while ( max_width > form->cursor + 1 &&
            *(string + form->cursor) != '\0' &&
            !isspace(*(string + form->cursor)) )
        form->cursor++;
    form_draw_cursor(form);
    return;

  MoveWordLeft:
    form_draw_cursor(form);
    /* Skip all blanks */
    while ( form->cursor > 0 &&
            isspace(*(string + form->cursor)) )
        form->cursor--;
    /* Skip all non-blanks */
    while ( form->cursor > 0 &&
            !isspace(*(string + form->cursor)) )
        form->cursor--;
    form_draw_cursor(form);
    return;

  NextField:
    form_draw_cursor(form); /* do it before changing form->curr_field */
    form->curr_field++;
    if ( form->curr_field >= NUM_FIELDS ) { /* moving off last field */
        /* If this is the last tuple in the list and all fields are empty, do
         * not do anything--just leave the cursor in the last field */
        if ( !form->curr_tuple->next && tuple_is_empty(form->curr_tuple) ) {
            form->curr_field = NUM_FIELDS-1;
            form->cursor     = 0; /* should be 0 anyway, good old paranoia */
        } else {
            /* Move on to the next tuple */
            if ( form->curr_tuple->next )
                form->curr_tuple = form->curr_tuple->next;
            else { /* no more tuples, allocate new one at the end */
                form->curr_tuple->next = tuple_alloc();
                form->curr_tuple->next->prev = form->curr_tuple;
                form->curr_tuple = form->curr_tuple->next;
            }
            form->cursor     = 0;
            /* Update screen and backing store */
            for ( form->curr_field = 0;
                  form->curr_field < NUM_FIELDS; form->curr_field++ )
                form_update_curr_field(form);
            form->curr_field = 0;
            /* Remove the previous tuple, if it's empty */
            tuple_maybe_unlink( &form->list_head, form->curr_tuple->prev );
        }
    } else {
        /* Move on to the next field, and make sure that cursor is not beyond
         * the end of string or field there */
        field = &form->fields[form->curr_field];
        max_width = strlen(form->curr_tuple->fields[form->curr_field]);
        if ( form->cursor > max_width )
            form->cursor = max_width;
        if ( form->cursor >= field->width )
            form->cursor = field->width - 1;
    }

    form_draw_cursor(form);
    return;

  PrevField:
    form_draw_cursor(form); /* do it before changing form->curr_field */
    form->curr_field--;
    if ( form->curr_field < 0 ) { /* moving off first field */
        if ( form->curr_tuple->prev ) {
            /* Move on to the previous tuple */
            form->curr_tuple = form->curr_tuple->prev;
            form->cursor     = 0;
            /* Update screen and backing store */
            for ( form->curr_field = 0;
                  form->curr_field < NUM_FIELDS; form->curr_field++ )
                form_update_curr_field(form);
            form->curr_field = NUM_FIELDS-1;
            /* Remove the next tuple, if it's empty */
            tuple_maybe_unlink( &form->list_head, form->curr_tuple->next );
        } else
            form->curr_field = 0;
    } else {
        /* Move on to the previous field, and make sure that cursor is not
         * beyond the end of string or field there */
        field = &form->fields[form->curr_field];
        max_width = strlen(form->curr_tuple->fields[form->curr_field]);
        if ( form->cursor > max_width )
            form->cursor = max_width;
        if ( form->cursor >= field->width )
            form->cursor = field->width - 1;
    }

    form_draw_cursor(form);
    return;

  DelChar:
    /* Shift all chars right of current position one char space to the left */
    form_draw_cursor(form);
    goto ShiftLeft;
    return;

  BSChar:
    if ( form->cursor > 0 ) {
        form_draw_cursor(form);
        form->cursor--;
        goto ShiftLeft;
    }
    return;

  ShiftLeft:
    string += form->cursor;
    while ( *string )
        *string++ = *(string + 1);
    form_update_curr_field(form);
    form_draw_cursor(form);
    return;

  KillToEnd:
    *(string + form->cursor) = '\0';
    form_draw_cursor(form);
    form_update_curr_field(form);
    form_draw_cursor(form);
    return;

  KillToStart:
    /* Shift all chars right of current position to the beginning of the string */
    string += form->cursor;
    while ( *string )
        *(string - form->cursor) = *string++;
    *(string - form->cursor) = '\0';
    form_draw_cursor(form);
    form->cursor = 0;
    form_update_curr_field(form);
    form_draw_cursor(form);
    return;

  DelWord:
    string += form->cursor;
    if ( *string ) {
        char *ptr = string;
        /* Skip non-word-constituent characters */
        while ( *ptr && !isalnum(*ptr) )
            ptr++;
        /* Skip word-constituent characters */
        while ( isalnum(*ptr) )
            ptr++;
        /* Shift the rest of the string, erasing the word */
        while ( *ptr )
            *string++ = *ptr++;
        *string = '\0';
        form_draw_cursor(form);
        form_update_curr_field(form);
        form_draw_cursor(form);
    }
    return;

  BSWord:
    if ( *string && form->cursor > 0 ) {
        char *ptr = string + form->cursor - 1; /* form->cursor is > 0 */
        int   cursor;
        /* Skip non-word-constituent characters */
        while ( ptr > string && !isalnum(*ptr) )
            ptr--;
        /* Skip word-constituent characters */
        if ( ptr > string ) {
            while ( ptr > string && isalnum(*ptr) )
                ptr--;
            if ( !isalnum(*ptr) )
                ptr++; /* put back the non-word-constituent character */
        }
        cursor = ptr - string;
        /* Shift the rest of the string, erasing the word */
        string += form->cursor;
        while ( *string )
            *ptr++ = *string++;
        *ptr = '\0';
        form_draw_cursor(form);
        form->cursor = cursor;
        form_update_curr_field(form);
        form_draw_cursor(form);
    }
    return;
}


Form *form_init( GdkDrawable *screen, GdkDrawable *pixmap,
                 GdkGC *fg_gc, GdkGC *bg_gc, Tuple *list_head )
{
    int i;
    Form *form;

    assert(fg_gc);
    assert(bg_gc);

    form = (Form *) xmalloc(sizeof(Form));
    if ( !form )
        return NULL;

    form->screen     = screen;
    form->pixmap     = pixmap;
    form->fg_gc      = fg_gc;
    form->bg_gc      = bg_gc;
    form->position.x = FORM_X;
    form->position.y = FORM_Y_PAD;
    form->list_head  = list_head;
    form->curr_tuple = list_head;
    form->curr_field = 0;
    form->cursor     = 0;

    form->fields = (FormField *) xmalloc( sizeof(FormField) * NUM_FIELDS );
    for ( i = 0; i < NUM_FIELDS; i++ ) {
        FormField *field = form->fields + i;
        field->position.x = 0;
        field->position.y = i * (font_char_height + FORM_SPACING);
        field->width = (field_dsc + i)->len;
    }

    if ( pixmap )
        form_display(form);

    form_draw_cursor( form );

    return form;
}



/*
 * Local Variables:
 * compile-command:  "make -k && ./bs -i sample.data"
 * End:
 */
