#include "bs.h"
#include <getopt.h>
#include <ctype.h>

/*
 * Global stuff
 */

GdkFont *font = NULL;
int font_char_width = 0, font_char_height = 0;
int FORM_X;
char *progname = NULL;
const char *errMemory = "%s: error allocating memory\n";

#define FIELD_LONGEST_NAME          4
FieldDsc field_dsc[NUM_FIELDS] = {
    { FLEN_ALBUM,        "Album",        NULL },
    { FLEN_TITLE,        "Title",        NULL },
    { FLEN_LENGTH,       "Length",       filter_time },
    { FLEN_AUTHOR,       "Author",       NULL },
    { FLEN_PERFORMED_BY, "Performed by", NULL },
    { FLEN_YEAR,         "Year",         isdigit }
};

Tuple *tuples_list = NULL;



/*
 * Local vars
 */
static GdkPixmap *backing_store = NULL;
static GtkWidget *main_window   = NULL;
static Form *form               = NULL;
static char *fname_in           = NULL;
static char *fname_out          = NULL;
static FILE *stream_in          = NULL;
static FILE *stream_out         = NULL;
static char *errOpenFile = "%s: error opening file `%s' for %s\n";

/*
 * Main window events and signals
 */

/* Close the application */
static void main_destroy(GtkWidget *widget, gpointer data)
{
    gtk_main_quit();

    return;
}

static gint key_event(GtkWidget *widget, GdkEventKey *event, gpointer data)
{
/*     g_print("0x%x, 0x%x\n", event->state, event->keyval); */

    form_process_char( form, event->state, event->keyval, *event->string );

    return FALSE; /* pass on the event */
}

/*
 * Drawing area events and signals
 */

static gint drawing_area_expose_event(GtkWidget *widget, GdkEventExpose *event)
{
    /* Redraw clobbered area of backing pixmap area */
    gdk_draw_pixmap( widget->window,
                     widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                     backing_store,
                     event->area.x, event->area.y,
                     event->area.x, event->area.y,
                     event->area.width, event->area.height );

    return FALSE;
}

/*
 * In configure event we recreate the backing pixmap with new size
 */
static gint drawing_area_configure_event(GtkWidget *widget,
                                         GdkEventConfigure *event)
{
    gint width, height;

    if ( backing_store )
        gdk_pixmap_unref(backing_store);

    width  = widget->allocation.width;
    height = widget->allocation.height;
    backing_store = gdk_pixmap_new(main_window->window, width, height, -1);

    /* Draw the form */
    if ( form )
        /* This is not the first time, so pass on list head from form struct */
        form = form_init( (GdkDrawable *) widget->window, backing_store,
                          widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                          widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
                          form->list_head );
    else {
        /* This is the first time, so pass on tuples_list for list head */
        form = form_init( (GdkDrawable *) widget->window, backing_store,
                          widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                          widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
                          tuples_list );
        /* From now on we don't know that this is still a valid list head */
        tuples_list = NULL;
    }

    return TRUE;
}


static int parse_options(int argc, char **argv);

int main(int argc, char **argv)
{
    int i;
    GtkWidget *drawing_area;

    /* GTK initialization */
    gtk_init(&argc, &argv);
    gtk_rc_parse("bsrc");

    stream_in  = stdin;
    stream_out = stdout;

    /* Parse command line */
    progname = *argv;
    i = parse_options(argc, argv);
    if ( i == 0 ) gtk_exit(-1);
    else if ( i == 1 ) gtk_exit(0);

    /* Open and read input file */
    if ( fname_in ) {
        stream_in = fopen(fname_in, "r");
        if ( !stream_in ) {
            fprintf(stderr, errOpenFile, progname, fname_in, "reading");
            gtk_exit(1);
        }
    } else
        stream_in = stdin;
    tuples_list = read_data(stream_in, fname_in);
    if ( !tuples_list )
        /* File is empty */
        tuples_list = tuple_alloc();

    /* Open output file */
    if ( fname_out ) {
        stream_out = fopen(fname_out, "w");
        if ( !stream_out ) {
            fprintf(stderr, errOpenFile, progname, fname_out, "writing");
            gtk_exit(1);
        }
    } else
        stream_out = stdout;

    /*
     * Main window
     */
    main_window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
    gtk_widget_set_events( main_window,
                           GDK_KEY_PRESS_MASK );
    gtk_widget_realize(main_window);
    /* Can both shrik and grow */
    gtk_window_set_policy( GTK_WINDOW(main_window), TRUE, TRUE, FALSE );
    /* Make it close on destroy signal */
    gtk_signal_connect( GTK_OBJECT(main_window), "destroy",
                        GTK_SIGNAL_FUNC(main_destroy), NULL );

    /*
     * Drawing area inside the main window
     */
    drawing_area = gtk_drawing_area_new();
    gtk_container_add( GTK_CONTAINER(main_window), drawing_area );
    gtk_signal_connect( GTK_OBJECT (main_window), "key_press_event",
                        GTK_SIGNAL_FUNC(key_event), (gpointer) drawing_area );
    gtk_signal_connect( GTK_OBJECT(drawing_area), "expose_event",
                        GTK_SIGNAL_FUNC(drawing_area_expose_event), NULL );
    gtk_signal_connect( GTK_OBJECT(drawing_area), "configure_event",
                        GTK_SIGNAL_FUNC(drawing_area_configure_event), NULL );
    font = drawing_area->style->font;
    font_char_width  = gdk_char_width(font, 'O');
    font_char_height = gdk_char_height(font, 'O');
    FORM_X = 2*FORM_X_PAD + gdk_string_width(font, field_dsc[FIELD_LONGEST_NAME].name);
    gtk_widget_set_usize(
        main_window,
        MAX_FLEN*font_char_width + FORM_X + FORM_X_PAD,
        NUM_FIELDS*font_char_height + (NUM_FIELDS-1)*FORM_SPACING + 2*FORM_Y_PAD );

    /* Show all widgets */
    gtk_widget_show(drawing_area);
    gtk_widget_show(main_window);

    /* Wait for events from X server */
    gtk_main();

    tuple_maybe_unlink(&form->list_head, form->curr_tuple);
    if ( !tuple_is_empty(form->curr_tuple) ) /* might happen if the user deleted
                                              * all tuples */
        write_data(form->list_head, stream_out);

    /* Close the streams */
    if ( stream_in != stdin )
        fclose(stream_in);
    if ( stream_out != stdout )
        fclose(stream_out);

    return 0;
}


static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-i <fname>] [-o <fname>]\n\
\n\
-h            display this help message\n\
-i <fname>    file to read data from, `-' for stdin (default: stdin)\n\
-o <fname>    file to write data to, `-' for stdout (default: stdout)\n\
\n\
Note: the usual EMACS binding are supported when editing text.\n\
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":i:o:h")) != -1 )
        switch ( c ) {
        case 'i':
            if ( *optarg == '-' && *(optarg+1) =='\0' )
                fname_in = NULL;
            else
                fname_in = optarg;
            break;
        case 'o':
            if ( *optarg == '-' && *(optarg+1) =='\0' )
                fname_out = NULL;
            else
                fname_out = optarg;
            break;
        case 'h':
            printf(usage, progname);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'i':
                msg = "input file name";
                break;
            case 'o':
                msg = "output file name";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    return -1;
}



/*
 * Local Variables:
 * compile-command:  "make -k && ./bs -i sample.data"
 * End:
 */
