#include "bs.h"
#include <ctype.h>

static char *errUnexpEOF = "%s: unexpected end of file while reading file `%s'\n";
static char *errFieldLen = "%s: error reading file `%s'--field `%s' too long (> %d chars)\n";

void *xmalloc(int size)
{
    void *ptr;

    if ( !(ptr = malloc(size)) ) {
        fprintf(stderr, errMemory, progname);
        exit(1);
    }

    return ptr;
}


int filter_time(int C)
{
    return isdigit(C) || C == ':';
}

/* Sets up a new tuple, allocating buffers for each of the field.  next and
 * prev members are set to NULL */
Tuple *tuple_alloc(void)
{
    Tuple *tuple = xmalloc(sizeof(Tuple));
    int i;

    tuple->next = NULL;
    tuple->prev = NULL;
    for ( i = 0; i < NUM_FIELDS; i++ ) {
        tuple->fields[i] = (char *) xmalloc(field_dsc[i].len + 1);
        *tuple->fields[i] = '\0';
    }

    return tuple;
}


int tuple_is_empty(Tuple *tuple)
{
    int i;

    assert(tuple);

    /* See if any of the fields is not empty */
    for ( i = 0; i < NUM_FIELDS; i++ )
        if ( strlen(tuple->fields[i]) )
            return 0;

    return 1;
}


/* Unlinks the tuple from linked list if it is not the only one in the list and
 * all its fields are empty.  If head tuple is freed, stores new head in *head.
 * Returns 1 if the tuple is unlinked, 0 if not */
int tuple_maybe_unlink(Tuple **head, Tuple *tuple)
{
    int i;

    assert(tuple);

    /* Do not unlink the tuple if it is the only one in the list or not empty */
    if ( (tuple->next || tuple->prev) && tuple_is_empty(tuple) ) {
        /* Free the mallocs and unlink the tuple */
        if ( tuple->prev )
            tuple->prev->next = tuple->next;
        else {
            /* Got a new head */
            assert(tuple == *head);
            *head = tuple->next;
        }
        if ( tuple->next )
            tuple->next->prev = tuple->prev;
        for ( i = 0; i < NUM_FIELDS; i++ )
            free(tuple->fields[i]);
        free(tuple);

        return 1;
    }

    return 0;
}


/* fname is for reporting purposes only */
Tuple *read_data(FILE *fstream, char *fname)
{
    Tuple *head = NULL, *curr;
    int i, j, c;

    assert(fstream);

    while ( 1 ) {
        /* Skip blanks */
        while ( isspace((c = getc(fstream))) )
            ;
        if ( c == EOF )
            break;
        ungetc(c, fstream);
        /* Allocate new tuple and add it to the end of (doubly linked) list */
        if ( !head )
            curr = head = tuple_alloc();
        else {
            curr->next = tuple_alloc();
            curr->next->prev = curr;
            curr = curr->next;
        }
        /* Read each of the fields (separated by '\n') */
        for ( i = 0; i < NUM_FIELDS; i++ ) {
            int max_len = field_dsc[i].len;
            char *str = curr->fields[i];

            for ( j=0;
                  (c = getc(fstream)) != EOF && c != '\n' && j < max_len;
                  j++ )
                *(str + j) = (char) c;
            if ( c == EOF ) { /* should not end in the middle of a tuple */
                fprintf(stderr, errUnexpEOF, progname, fname);
                gtk_exit(1);
            }
            if ( j == max_len && c != '\n' ) { /* field too long */
                fprintf(stderr, errFieldLen, progname, fname, field_dsc[i].name, max_len);
                gtk_exit(1);
            }
            *(str + j) = '\0';
        }
    }

    return head;
}


void write_data(Tuple *head, FILE *fstream)
{
    int i;

    assert(head);

    while ( head ) {
        for ( i = 0; i < NUM_FIELDS; i++ )
            fprintf(fstream, "%s\n", head->fields[i]);
        fputc('\n', fstream);
        head = head->next;
    }

    return;
}



/*
 * Local Variables:
 * compile-command:  "make -k && ./bs -i sample.data"
 * End:
 */
