#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>
#include <math.h>

#include "wrapper.h"

char *errUnknownOption = "%s: error--unknown option `-%c'\n";
char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
char *errDelayRange    = "%s: error--delay must be an integer in the range [ 0, 10e9 )\n";
char *errGranularityRange = "%s: error--granularity must be an integer of 1, 2, 3, or 6\n";
char *errInteger       = "%s: error--%s must be an integer\n";
char *errArgumentReq   = "%s: error--option `%s' requires an argument (%s)\n";
char *errFont          = "%s: error--cannot load font\n";

char *usehelp = "Try `%s -h' for help\n";
char *usage   = "\
Usage: %s [-h] [-d nsec]\n\n\
-h       display this help message\n\
-g       granularity--how many times to tick per second, can be 1, 2, 3, or 6\n\
         (default: %d)\n\
-d nsec  delay each stroke nsec nanoseconds, must be an integer in the range\n\
         [ 0, 10e9 ) (default: draw real-time)\n\n";

int  granularity = 6;
int  use_delay = 0;
long delay;
struct timespec rtime;

int  fontheight, scrwidth, scrheight, colordepth;

typedef struct _point {
    int x;
    int y;
} point, *ppoint;

point hhand[] = {  { 0, -10}, {-10,   0}, { 0, 120}, { 10, 0}  };
point mhand[] = {  { 0, -10}, { -5,   0}, { 0, 175}, {  5, 0}  };
point shand[] = {  { 0, -10}, {  0, 175}  };
/* point hhand[] = {  { 4, -10}, {-4, -10}, { -4, 120}, { 4, 120}  }; */
/* point mhand[] = {  { 2,  -6}, {-2,  -6}, { -2, 175}, { 2, 175}  }; */
/* point shand[] = {  { 0,  -4}, { 0, 175}  }; */

#define HHAND_LEN           sizeof(hhand) / sizeof(point)
#define MHAND_LEN           sizeof(mhand) / sizeof(point)
#define SHAND_LEN           sizeof(shand) / sizeof(point)
#define CLOCK_RADIUS        200 /* real-world radius */


void draw_face(int center_x, int center_y, int radius);
void run_clock(int center_x, int center_y, int radius);
void draw_hand(const ppoint hand, int len, int c, int r, int a, int b,
               double ex, double ey);
#ifdef __LINUX__
int parse_options(int argc, char **argv);
#endif /* __LINUX__ */


int main(int argc, char **argv)
{
    char c;
    int  i;

#ifdef __LINUX__
    /* Process command line arguments */
    i = parse_options(argc, argv);
    if ( i == 0 )
        return -1;
    else if ( i == 1 )
        return 0;
#endif /* __LINUX__ */

    /* Setup graphics */
    VGA_INIT;
    fontheight = 8;
    if ( !loadfont(8, fontheight, "font.8") ) {
        VGA_CLOSE;
        fprintf(stderr, errFont, *argv);
        return -1;
    }
    scrwidth   = vga_getxdim();
    scrheight  = vga_getydim();
    colordepth = vga_getcolors();

    /* Set delay */
    if ( use_delay ) {
        rtime.tv_sec  = 0;
        rtime.tv_nsec = delay;
    } else {
        if ( granularity == 1 ) {
            rtime.tv_sec  = (time_t) 1;
            rtime.tv_nsec = (long) 0;
        }
        else {
            rtime.tv_sec  = (time_t) 0;
            rtime.tv_nsec = (long) 1000000000l / granularity; /* < 1 bil */
        }
    }

    draw_face(scrwidth/2, scrheight/2, scrheight/3);

    run_clock(scrwidth/2, scrheight/2, scrheight/3);

    /* Wait for a key and bail out */
/*     scanf("%c", &c); */
    VGA_CLOSE;

    return 0;
}


/* Convert hour, minute and second values into degrees of rotation */
#define HOU_TO_H(hou, min)      (hou)*30 + (min)/2 /* min*6/12 */
#define MIN_TO_M(min, sec)      (min)*6  + (sec)   /* sec*60/60 */
#define SEC_TO_S(sec)           (sec)*6

void run_clock(int center_x, int center_y, int radius)
{
    unsigned hou, min, h, m, s, last_h, last_m, last_s, step;
    double ex, ey;
    struct tm *timel;
    time_t timet;

    ex = ((double) radius)/CLOCK_RADIUS;
    ey = ((double) radius)/CLOCK_RADIUS;
    if ( ex < ey ) ey = ex; else ex = ey;

    timet  = time(NULL);
    timel  = localtime(&timet);

    hou = timel->tm_hour % 12;
    min = timel->tm_min;
    s   = SEC_TO_S(timel->tm_sec);
    last_h = last_m = last_s = 0;
    step = 6 / granularity;

    while ( 1 ) {
        hou %= 12;
        for ( ; min < 60; min++ ) {
            for ( ; s < 360; s += step ) {
                draw_hand(hhand, HHAND_LEN, 0, -last_h, center_x, center_y, ex, ey);
                draw_hand(mhand, MHAND_LEN, 0, -last_m, center_x, center_y, ex, ey);
                draw_hand(shand, SHAND_LEN, 0, -last_s, center_x, center_y, ex, ey);

                m = MIN_TO_M(min, s/60);
                h = HOU_TO_H(hou, min);
                draw_hand(hhand, HHAND_LEN, 7, -h, center_x, center_y, ex, ey);
                draw_hand(mhand, MHAND_LEN, 7, -m, center_x, center_y, ex, ey);
                draw_hand(shand, SHAND_LEN, 6, -s, center_x, center_y, ex, ey);
                last_h = h;
                last_m = m;
                last_s = s;
                if ( rtime.tv_sec || rtime.tv_nsec )
                    nanosleep(&rtime, NULL);
/*                 vga_waitretrace(); */
            }
            s = 0;
        }
        min = MIN_TO_M(0, 0);
        hou++;
    }

    return;
}


static char num[3];


void draw_face(int center_x, int center_y, int radius)
{
    int x, y, i;

    center_y -= fontheight/2;
    radius   -= fontheight/2;

    vga_setcolor(7);
    for ( i = 1; i <= 60; i++ ) {
        x = radius * cos(i*6/180.*M_PI - M_PI_2) + center_x;
        y = radius * sin(i*6/180.*M_PI - M_PI_2) + center_y;
        if ( i % 5 )
            *num = '.', *(num + 1) = '\0';
        else
            sprintf(num, "%d",  i/5);
        textout(x, y, num);
    }

    return;
}


/*
 * We need to scale the hand in directions of coordinate axes with ratios
 * (ex, ey), rotate it (pi + r) degrees, and then move it by vector (a, b).
 * The entire transformation can be expressed as a 3x3 matrix resulting from
 * multiplication of three 3x3 matrices R(pi + r), T(a, b) and E(ex,ey) for
 * rotation, moving and scaling:
 *
 *  phi = pi + r,
 *
 *         ( cos(phi) sin(phi) 0 )          ( 1 0 a )            ( ex 0  0 )
 *  R(phi)=(-sin(phi) cos(phi) 0 ), T(a, b)=( 0 1 b ), E(ex, ey)=( 0  ey 0 )
 *         (     0        0    1 )          ( 0 0 1 )            ( 0  0  1 )
 *
 * So the new coordinates are:
 *
 *  ( X )   ( 1  0  a )   (  cos(phi)  sin(phi)    0 )   ( ex 0  0 )   ( x )
 *  ( Y ) = ( 0  1  b ) x ( -sin(phi)  cos(phi)    0 ) x ( 0  ey 0 ) x ( y ) =
 *  ( 1 )   ( 0  0  1 )   (      0         0       1 )   ( 0  0  1 )   ( 1 )
 *
 *          (  ex * ( x*cos(phi) + y*sin(phi) )  +  a  )
 *        = (  ey * (-x*sin(phi) + y*cos(phi) )  +  b  )
 *          (                    1                     )
 */
#define TRANSFORM_X          (ex * ( new_x * cos_phi + new_y * sin_phi )  +  a)
#define TRANSFORM_Y          (ey * (-new_x * sin_phi + new_y * cos_phi )  +  b)

void draw_hand(const ppoint hand, int len, int c, int r, int a, int b,
               double ex, double ey)
{
    int i, last_x, last_y, new_x, new_y, x, y;
    double phi, cos_phi, sin_phi;

    phi = M_PI + r / 180. * M_PI;
    cos_phi = cos(phi);
    sin_phi = sin(phi);

    new_x  = (hand+len-1)->x;
    new_y  = (hand+len-1)->y;
    last_x = TRANSFORM_X;
    last_y = TRANSFORM_Y;

    vga_setcolor(c);
    for ( i = 0; i < len; i++ ) {
        new_x = (hand+i)->x;
        new_y = (hand+i)->y;
        x = TRANSFORM_X;
        y = TRANSFORM_Y;
        vga_drawline(last_x, last_y, x, y);
        last_x = x;
        last_y = y;
    }

    return;
}


#ifdef __LINUX__
/* Return codes:
 * 0    - error, exit(-1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
int parse_options(int argc, char **argv)
{
    char c, *ptr;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":g:d:h")) != -1 )
        switch ( c ) {
        case 'g':
            granularity = (int) strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "granularity");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            if ( granularity != 1 && granularity != 2 &&
                 granularity != 3 && granularity != 6 ) {
                fprintf(stderr, errGranularityRange, *argv);
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            break;
        case 'd':
            delay = strtol(optarg, &ptr, 0);
            if ( ptr == optarg ) {
                fprintf(stderr, errInteger, *argv, "delay");
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            if ( delay < 0 || delay >= 1000000000 ) {
                fprintf(stderr, errDelayRange, *argv);
                fprintf(stderr, usehelp, *argv);
                return 0;
            }
            use_delay = 1;
            break;
        case 'h':
            printf(usage, *argv, granularity);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, *argv, optopt);
                fprintf(stderr, usehelp, *argv);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, *argv, optopt);
                fprintf(stderr, usehelp, *argv);
            }
            return 0;
            break;
        case ':': /* missing option argument */
            switch ( optopt ) {
            case 'd':
                fprintf(stderr, errArgumentReq, *argv, "-d", "delay");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            case 'g':
                fprintf(stderr, errArgumentReq, *argv, "-g", "granularity");
                fprintf(stderr, usehelp, *argv);
                return 0;
                break;
            default:
                abort(); /* should never get here */
                break;
            }
        default:
            abort(); /* should never happen */
        }

    return -1;
}
#endif /* __LINUX__ */



/*
 * Local variables:
 * compile-command: "make -k"
 * End:
 */
