#include "clock.h"


static void draw_face (GtkWidget *widget);
static void draw_hand (GtkWidget *widget, GdkGC *gc,
                       const GdkPoint *hand, gint len, gint r);
static gint clock_tick(GtkWidget *widget);
static void draw_hands(GtkWidget *widget, GdkGC **shand_gc, gboolean erase);




static GdkPoint hhand[] = { { 0, -10}, {-10,   0}, { 0, 120}, { 10, 0}, {0, -10}, {0, 120} };
static GdkPoint mhand[] = { { 0, -10}, { -5,   0}, { 0, 175}, {  5, 0}, {0, -10}, {0, 175} };
static GdkPoint shand[] = { { 0, -10}, {  0, 175} };
/* static GdkPoint hhand[] = { { 4, -10}, {-4, -10}, { -4, 120}, { 4, 120}, {4, -10} }; */
/* static GdkPoint mhand[] = { { 2,  -6}, {-2,  -6}, { -2, 175}, { 2, 175}, {2,  -6} }; */
/* static GdkPoint shand[] = { { 0,  -4}, { 0, 175} }; */
#define HHAND_LEN           ( sizeof(hhand) / sizeof(hhand[0]) )
#define MHAND_LEN           ( sizeof(mhand) / sizeof(mhand[0]) )
#define SHAND_LEN           ( sizeof(shand) / sizeof(shand[0]) )
#define HANDS_RADIUS        175 /* real-world maximum radius */

#define SEC_RES_COUNT       ( sizeof(sec_res_val) / sizeof(sec_res_val[0]) )
#define MIN_RES_COUNT       ( sizeof(min_res_val) / sizeof(min_res_val[0]) )
#define HOU_RES_COUNT       ( sizeof(hou_res_val) / sizeof(hou_res_val[0]) )

/* Movement resolutions--hands can "jerk" by this many 1/10 of degrees
 * ("granularity units").  Must be a divisor of 60 for minutes and seconds and
 * of 300 for hours (which are one-unit "jerks") */
const gint sec_res_val[] = {1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60};
const gint min_res_val[] = {1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60};
const gint hou_res_val[] = {1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 25, 30, 50, 60, 75, 100, 150, 300};
const gint sec_res_count = SEC_RES_COUNT;
const gint min_res_count = MIN_RES_COUNT;
const gint hou_res_count = HOU_RES_COUNT;

/* Clock's hands movement resolutions in 1/10 of degree.  Initialized in
 * clock_init().  Use one of the above *_res_val arrays to set these */
static gint sec_res_no = 0, min_res_no = 0, hou_res_no = 0;
/* Number of granularity units which hands have already travelled.  Initialized
 * in clock_init() */
static gint sec_tick = 0, min_tick = 0, hou_tick = 0;

static GdkPixmap *pixmap = NULL;      /* backing pixmap */
static GdkGC     *shand_gc[5];        /* GC's for seconds hand */
static GtkWidget *ClockDrawingArea = NULL;
static guint      timer_tag = 0;      /* uninitialized */
static guint32    timer_interval = 0; /* initialized in clock_init() */
static gboolean   keep_aspect_ratio = 0;
static gboolean   run_real_time = FALSE;
static gboolean   clock_initialized = FALSE;



static void destroy(GtkWidget *widget)
{
    gtk_timeout_remove( timer_tag );

    clock_initialized = FALSE;

    return;
}

static gint expose_event(GtkWidget *widget, GdkEventExpose *event)
{
    /* Redraw clobbered area of backing pixmap area */
    gdk_draw_pixmap(widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    pixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height);

    /* Draw clock hands */
    draw_hands(widget, shand_gc, FALSE);

    return FALSE;
}

/*
 * In configure event we recreate the backing pixmap with new size.  We draw
 * clock's face in backing pixmap, but hands are redrawn from scratch in
 * expose_event.  This is to avoid outputting entire backing pixmap each clock
 * tick.
 */
static gint configure_event(GtkWidget *widget, GdkEventConfigure *event)
{
    gint width, height;

    if ( pixmap )
        gdk_pixmap_unref(pixmap);

    width  = widget->allocation.width;
    height = widget->allocation.height;
    pixmap = gdk_pixmap_new(widget->window, width, height, -1);

    /* Blank out the rectangle */
    gdk_draw_rectangle( pixmap,
                        widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
                        TRUE,
                        0, 0,
                        widget->allocation.width,
                        widget->allocation.height);
    /* Draw new clock  face */
    draw_face(widget);

    return TRUE;
}



static void draw_face(GtkWidget *widget)
{
    GdkPoint center, point, radius;
    gint     i, textheight, textwidth, notch_radius;
    gchar    textbuf[3];

    /* Find out height of digits */
    *textbuf = '0', *(textbuf+1) = '\0';
    textheight = gdk_string_height(widget->style->font, textbuf);
    /* Calculate parameters of clock face */
    center.x  = (widget->allocation.width + 1)/2;
    center.y  = (widget->allocation.height + 1)/2;
    if ( keep_aspect_ratio ) {
        /* reserve space for hour numbers */
        radius.x = radius.y = \
            (center.x < center.y ? center.x : center.y) - 1 - (textheight+1)/2;
    } else {
        radius.x = center.x - 1 - (textheight+1)/2;
        radius.y = center.y - 1 - (textheight+1)/2;
    }

    notch_radius = textheight/10;
    if ( notch_radius < 1 )
        notch_radius = 1;

    for ( i = 1; i <= 60; i++ ) {
        point.x = radius.x * cos(i*6/180.*M_PI - M_PI_2) + center.x;
        point.y = radius.y * sin(i*6/180.*M_PI - M_PI_2) + center.y;
        if ( i % 5 ) { /* Draw notch */
            gdk_draw_arc(pixmap,
                         widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                         FALSE,
                         point.x - notch_radius, point.y - notch_radius,
                         2*notch_radius, 2*notch_radius, 0, 360*64);
        } else {       /* Draw number */
            sprintf(textbuf, "%d",  i/5);
            textwidth = gdk_string_width(widget->style->font, textbuf);
            point.x -= textwidth/2;
            point.y += textheight/2;
            gdk_draw_string(pixmap,
                            widget->style->font,
                            widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                            point.x, point.y, textbuf);
        }
    }

    return;
}


static void draw_hands(GtkWidget *widget, GdkGC **shand_gc, gboolean erase)
{
    if ( erase ) {
        draw_hand( widget, widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
                   hhand, HHAND_LEN, hou_tick);
        draw_hand( widget, widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
                   mhand, MHAND_LEN, min_tick);
        draw_hand( widget, widget->style->bg_gc[GTK_WIDGET_STATE(widget)],
                   shand, SHAND_LEN, sec_tick);
    } else {
        draw_hand( widget, widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                   hhand, HHAND_LEN, hou_tick);
        draw_hand( widget, widget->style->fg_gc[GTK_WIDGET_STATE(widget)],
                   mhand, MHAND_LEN, min_tick);
        draw_hand( widget, *(shand_gc + GTK_WIDGET_STATE(widget)),
                   shand, SHAND_LEN, sec_tick);
    }
    return;
}

/*
 * We need to scale the hand in directions of coordinate axes with ratios
 * (ex,ey), rotate it (pi + (-r)/10) degrees, and then move it by vector (a,b).
 * The entire transformation can be expressed as a 3x3 matrix resulting from
 * multiplication of three 3x3 matrices R(pi + (-r)/10), T(a,b) and E(ex,ey) for
 * rotation, moving and scaling:
 *
 *  phi = pi + (-r) / 10,
 *
 *         ( cos(phi) sin(phi) 0 )          ( 1 0 a )            ( ex 0  0 )
 *  R(phi)=(-sin(phi) cos(phi) 0 ), T(a, b)=( 0 1 b ), E(ex, ey)=( 0  ey 0 )
 *         (     0        0    1 )          ( 0 0 1 )            ( 0  0  1 )
 *
 * So the new coordinates are:
 *
 *  ( X )   ( 1  0  a )   (  cos(phi)  sin(phi)    0 )   ( ex 0  0 )   ( x )
 *  ( Y ) = ( 0  1  b ) x ( -sin(phi)  cos(phi)    0 ) x ( 0  ey 0 ) x ( y ) =
 *  ( 1 )   ( 0  0  1 )   (      0         0       1 )   ( 0  0  1 )   ( 1 )
 *
 *          (  ex * ( x*cos(phi) + y*sin(phi) )  +  a  )
 *        = (  ey * (-x*sin(phi) + y*cos(phi) )  +  b  )
 *          (                    1                     )
 */
#define TRANSFORM_X          (ex * ( new.x * cos_phi + new.y * sin_phi )  +  a)
#define TRANSFORM_Y          (ey * (-new.x * sin_phi + new.y * cos_phi )  +  b)

static void draw_hand(GtkWidget *widget, GdkGC *gc,
               const GdkPoint *hand, gint len, gint r)
{
    gchar    textbuf[2];
    gint     i, a, b, radius, textheight;
    gdouble  ex, ey;
    GdkPoint last, new, curr;
    gdouble  phi, cos_phi, sin_phi;

    /* Find out height of digits */
    *textbuf = '0', *(textbuf+1) = '\0';
    textheight = gdk_string_height(widget->style->font, textbuf);
    a  = (widget->allocation.width + 1)/2;
    b  = (widget->allocation.height + 1)/2;
    if ( keep_aspect_ratio ) {
        radius  = (a < b ? a : b) - 1 - textheight * 1.4142135623730951;
        ex = ey = ((gdouble) radius) / HANDS_RADIUS;
    } else {
        ex = ((gdouble) a - 1 - textheight * 1.4142135623730951) / HANDS_RADIUS;
        ey = ((gdouble) b - 1 - textheight * 1.4142135623730951) / HANDS_RADIUS;
    }

    phi = M_PI - r / 1800. * M_PI;
    cos_phi = cos(phi);
    sin_phi = sin(phi);

    new.x  = hand->x;
    new.y  = hand->y;
    last.x = TRANSFORM_X;
    last.y = TRANSFORM_Y;

    for ( i = 1; i < len; i++ ) {
        new.x = (hand+i)->x;
        new.y = (hand+i)->y;
        curr.x = TRANSFORM_X;
        curr.y = TRANSFORM_Y;
        gdk_draw_line( widget->window, gc,  last.x, last.y, curr.x, curr.y);
        last = curr;
    }

    return;
}

static void set_ticks_from_time(gint hou, gint min, gint sec)
{
    gint sec_res = *(sec_res_val + sec_res_no);
    gint min_res = *(min_res_val + min_res_no);
    gint hou_res = *(hou_res_val + hou_res_no);

    sec_tick = sec * 3600/60;
    sec_tick = sec_tick - sec_tick % sec_res;

    min_tick = min * 3600/60 + sec; /* sec == sec_tick / (3600/60) */
    min_tick = min_tick - min_tick % min_res;

    hou_tick = hou * 3600/12 + min * 5; /* min == min_tick / (3600/60) */
    hou_tick = hou_tick - hou_tick % hou_res;

    return;
}

static gint clock_tick(GtkWidget *widget)
{
    gint new_sec_tick, new_min_tick, new_hou_tick;
    gint sec_res = *(sec_res_val + sec_res_no);
    gint min_res = *(min_res_val + min_res_no);
    gint hou_res = *(hou_res_val + hou_res_no);

    if ( run_real_time ) {
        clock_set_local_time();
        return TRUE;
    }

    new_min_tick = min_tick;
    new_hou_tick = hou_tick;

    new_sec_tick = (sec_tick + sec_res) % 3600;
    if ( new_sec_tick % (min_res * 60) == 0 ) {
        new_min_tick = (new_min_tick + min_res) % 3600;
        if ( new_min_tick % (hou_res * 12) == 0 )
            new_hou_tick = (new_hou_tick + hou_res) % 3600;
    }

    draw_hands(widget, shand_gc, TRUE);

    sec_tick = new_sec_tick;
    min_tick = new_min_tick;
    hou_tick = new_hou_tick;

    draw_hands(widget, shand_gc, FALSE);

    return TRUE;
}


/*
 *
 * Below are `public' functions
 *
 */


/* Change movement resolution of clock hands.  Pass indexes into *_res_val
 * arrays.  To leave resolution unchanged, pass negative index */
void clock_set_resolution_indexes(gint hou_res_NO, gint min_res_NO,
                                  gint sec_res_NO)
{
    gint new_sec_res_no, new_min_res_no, new_hou_res_no;

    assert( clock_initialized ); /* should not be called before clock_init() */

    new_sec_res_no = sec_res_no;
    new_min_res_no = min_res_no;
    new_hou_res_no = hou_res_no;

    /* Validate resolution indexes */
    if ( sec_res_NO >= 0 ) {
        if ( sec_res_NO >= SEC_RES_COUNT )
            sec_res_NO = SEC_RES_COUNT - 1;
        new_sec_res_no = sec_res_NO;
    }

    if ( min_res_NO >= 0 ) {
        if ( min_res_NO >= MIN_RES_COUNT )
            min_res_NO = MIN_RES_COUNT - 1;
        new_min_res_no = min_res_NO;
    }

    if ( hou_res_NO >= 0 ) {
        if ( hou_res_NO >= HOU_RES_COUNT )
            hou_res_NO = HOU_RES_COUNT - 1;
        new_hou_res_no = hou_res_NO;
    }

    /* Hide hands */
    draw_hands(ClockDrawingArea, shand_gc, TRUE);

    /* Set new resolution values */
    sec_res_no = new_sec_res_no;
    min_res_no = new_min_res_no;
    hou_res_no = new_hou_res_no;

    /* Recalculate ticks with new resolution values */
    set_ticks_from_time( hou_tick / (3600/12),
                         min_tick / (3600/60),
                         sec_tick / (3600/60) );

    draw_hands(ClockDrawingArea, shand_gc, FALSE);

    return;
}

/* Change movement resolution of clock hands.  Passed values will be adjusted to
 * the nearest possible values (according to *_res_val arrays).  To leave
 * resolution unchanged, pass non-positive value */
void clock_set_resolutions(gint hou_res, gint min_res, gint sec_res)
{
    gint i;

    assert( clock_initialized ); /* should not be called before clock_init() */

    /* Validate resolution values and translate them into indexes */
    if ( sec_res > 0 ) {
        for ( i = 0; i < SEC_RES_COUNT; i++ )
            if ( sec_res <= *(sec_res_val + i) ) {
                sec_res = i;
                break;
            }
        if ( i == SEC_RES_COUNT )
            sec_res = SEC_RES_COUNT - 1;
    } else
        sec_res = -1;

    if ( min_res > 0 ) {
        for ( i = 0; i < MIN_RES_COUNT; i++ )
            if ( min_res <= *(min_res_val + i) ) {
                min_res = i;
                break;
            }
        if ( i == MIN_RES_COUNT )
            min_res = MIN_RES_COUNT - 1;
    } else
        min_res = -1;

    if ( hou_res > 0 ) {
        for ( i = 0; i < HOU_RES_COUNT; i++ )
            if ( hou_res <= *(hou_res_val + i) ) {
                hou_res = i;
                break;
            }
        if ( i == HOU_RES_COUNT )
            hou_res = HOU_RES_COUNT - 1;
    } else
        hou_res = -1;

    /* Set the new resolution indexes */
    clock_set_resolution_indexes( hou_res, min_res, sec_res );

    return;
}

void clock_get_resolution_indexes(gint *hou, gint *min, gint *sec)
{
    if ( sec )
        *sec = sec_res_no;
    if ( min )
        *min = min_res_no;
    if ( hou )
        *hou = hou_res_no;

    return;
}

/* Change time displayed by the clock */
void clock_set_time(guint hou, guint min, guint sec)
{
    assert( clock_initialized ); /* should not be called before clock_init() */

    hou %= 12;
    min %= 60;
    sec %= 60;

    /* Hide hands */
    draw_hands(ClockDrawingArea, shand_gc, TRUE);

    /* Recalculate ticks with new time values */
    set_ticks_from_time(hou, min, sec);

    draw_hands(ClockDrawingArea, shand_gc, FALSE);

    return;
}

/* Get time displayed by the clock (for the parts which are not NULL) */
void clock_get_time(guint *hou, guint *min, guint *sec)
{
    if ( sec )
        *sec = sec_tick / 60;
    if ( min )
        *min = min_tick / 60;
    if ( hou )
        *hou = hou_tick / 300;

    return;
}

/* Change clock's display to local time */
void clock_set_local_time(void)
{
    struct tm *timel;
    time_t timet;

    assert( clock_initialized ); /* should not be called before clock_init() */

    timet  = time(NULL);
    timel  = localtime(&timet);

    clock_set_time(timel->tm_hour % 12, timel->tm_min, timel->tm_sec);

    return;
}

/* Change timer interval.  Value is in miliseconds and must be from 1 to 1000
 * inclusive.  Pass 0 to have interval calculated from movement resolutions to
 * produce approximately real-time movement */
void clock_set_timer_interval(guint32 interval)
{
    assert( clock_initialized ); /* should not be called before clock_init() */

    if ( !interval )
        interval = (guint32) ( 1000 / (60 / *(sec_res_val + sec_res_no)) );
    else if ( interval > 1000 )
        interval = 1000;

    timer_interval = interval;

    /* Restart timer if timer already existed; otherwise leave clock pausing */
    if ( timer_tag ) {
        gtk_timeout_remove(timer_tag);
        timer_tag = gtk_timeout_add( timer_interval, (GtkFunction) clock_tick,
                                     (gpointer) ClockDrawingArea );
    }

    return;
}

/* Run or stop the clock */
void clock_run(gboolean run)
{
    assert( clock_initialized ); /* should not be called before clock_init() */

    if ( run ) {
        if ( !timer_tag )
            timer_tag = gtk_timeout_add(timer_interval, (GtkFunction)clock_tick,
                                        (gpointer) ClockDrawingArea );
    } else
        if ( timer_tag ) {
            gtk_timeout_remove( timer_tag );
            timer_tag = 0;
        }

    return;
}

/* Pass TRUE to make clock remain round no matter how user resizes window; pass
 * FALSE to make clock break aspect ration and turn into ellipse if window's
 * width != window's height */
void clock_keep_aspect_ratio(gboolean keep)
{
    assert( clock_initialized ); /* should not be called before clock_init() */

    keep_aspect_ratio = keep;
    gtk_widget_queue_resize(ClockDrawingArea);
    gtk_widget_queue_draw(ClockDrawingArea);

    return;
}

/* If you pass TRUE to this function, the clock will tick every real-time
 * second, without any delay.  Note that in this mode the seconds hand's
 * movement resolution is not honored.  To restore the previous state, pass
 * FALSE to this function */
void clock_run_real_time(gboolean real_time)
{
    static guint32 tinterval;

    assert( clock_initialized ); /* should not be called before clock_init() */

    if ( real_time ) {           /* arrange the clock to run real-time */
        if ( !run_real_time ) {
            tinterval = timer_interval; /* remember current timer interval */
            clock_set_timer_interval(REAL_TIME_TIMER_INTERVAL);
            run_real_time = TRUE;
        }
    } else {                     /* restore the clock's state */
        if ( run_real_time ) {
            timer_interval = tinterval;
            clock_set_timer_interval(timer_interval);
            run_real_time = FALSE;
        }
    }

    return;
}

/*
 * Clock initialization.  Parameters:
 *     darea           drawing area widget to draw upon;
 *     interval        interval to pause between each seconds hand's tick
 *                     (default: set calculated from movement resolutions to
 *                     produce approximately real-time movement);
 *     hou, min, sec   time to set display to (default: local time);
 *     *_res_NO        array indexes to use when setting hands' resolutions.
 *
 * To force defaults, pass 0 for interval and negative values for hou and
 * *_res_NO parameters.
 *
 * clock_init() does not run the clock; start it explicitely with
 * clock_run(TRUE).
 *
 * Any time after initialization, you can change timer interval, clock hands
 * movement resolution, and displayed time by calling relevant functions
 */
gint clock_init( GtkDrawingArea *darea,
                 guint32 interval,
                 gint hou, gint min, gint sec,
                 gint hou_res_NO, gint min_res_NO, gint sec_res_NO )
{
    GdkColor *color;
    gint i;

    assert( !clock_initialized ); /* should only be called once */

    /*
     * Initialize parameters and variables
     */
    if ( hou_res_NO < 0 ) hou_res_NO     = DEFAULT_HOU_RESOLUTION_NO;
    if ( min_res_NO < 0 ) min_res_NO     = DEFAULT_MIN_RESOLUTION_NO;
    if ( sec_res_NO < 0 ) sec_res_NO     = DEFAULT_SEC_RESOLUTION_NO;
    ClockDrawingArea = GTK_WIDGET(darea);
    /* From now on `public' clock functions will work */
    clock_initialized = TRUE;

    /*
     * Seconds hand requires special GC's because it should be red
     */
    /* Create one GC for each possible state and copy default GC's into them */
    for ( i = 0; i < 5; i++ ) {
        shand_gc[i] = gdk_gc_new(ClockDrawingArea->window);
        gdk_gc_copy(shand_gc[i], ClockDrawingArea->style->light_gc[i]);
    }
    color = (GdkColor *) malloc( sizeof(GdkColor) );
    /*  Allocate dark red color for insensitive state's GC */
    color->red   = 0xa000;
    color->green = 0x6000;
    color->blue  = 0x6000;
    color->pixel = (0xa000ul << 16) + (0x6000ul << 8) + 0x6000;
    gdk_color_alloc(gtk_widget_get_colormap(ClockDrawingArea), color);
    gdk_gc_set_foreground(shand_gc[GTK_STATE_INSENSITIVE], color);
    /*  Allocate bright red color for other states' GC's */
    color->red   = 0xd000;
    color->green = 0x6000;
    color->blue  = 0x6000;
    color->pixel = (0xd000ul << 16) + (0x6000ul << 8) + 0x6000;
    gdk_color_alloc(gtk_widget_get_colormap(ClockDrawingArea), color);
    gdk_gc_set_foreground(shand_gc[GTK_STATE_NORMAL],   color);
    gdk_gc_set_foreground(shand_gc[GTK_STATE_ACTIVE],   color);
    gdk_gc_set_foreground(shand_gc[GTK_STATE_PRELIGHT], color);
    gdk_gc_set_foreground(shand_gc[GTK_STATE_SELECTED], color);
    /* Free the mallocs! */
    free(color);

    /*
     * Signals and events
     */
    gtk_signal_connect( GTK_OBJECT(ClockDrawingArea), "expose_event",
                        GTK_SIGNAL_FUNC(expose_event), NULL );
    gtk_signal_connect( GTK_OBJECT(ClockDrawingArea), "configure_event",
                        GTK_SIGNAL_FUNC(configure_event), NULL );
    gtk_signal_connect( GTK_OBJECT(ClockDrawingArea), "destroy",
                        GTK_SIGNAL_FUNC(destroy), NULL );

    /*
     * Time, timer and movement resolutions
     */
    clock_set_resolution_indexes(hou_res_NO,
                                 min_res_NO,
                                 sec_res_NO); /* should call
     * this before clock_set_timer_interval(), because the latter uses
     * sec_res_no if interval == 0 */
    clock_set_timer_interval(interval);
    if ( hou < 0 )
        clock_set_local_time();
    else
        clock_set_time(hou, min, sec);

    keep_aspect_ratio = TRUE;

    return 0;
}



/*
 * Local variables:
 * compile-command: "make -k && ./clockgtk"
 * End:
 */
