#include "threads.h"
#include <math.h>


/* Clock hands (real-world coordinates) */
static point hhand[] = {  { 0, -10}, {-10,   0}, { 0, 120}, { 10, 0}  };
static point mhand[] = {  { 0, -10}, { -5,   0}, { 0, 175}, {  5, 0}  };
static point shand[] = {  { 0, -10}, {  0, 175}  };
/* Number of points in arrays */
#define HHAND_LEN           sizeof(hhand) / sizeof(point)
#define MHAND_LEN           sizeof(mhand) / sizeof(point)
#define SHAND_LEN           sizeof(shand) / sizeof(point)
/* Real-world clock radius for which the hands were designed */
#define CLOCK_RADIUS        200



static void draw_face(unsigned char *bstore, int width,
                      point center, point radius);
static void draw_border(unsigned char *bstore, int width, int height);

static void draw_hand(unsigned char *bstore, int width,
                      const point *hand, int len, char color,
                      int angle, int a, int b, double ex, double ey);



/* Convert hour, minute and second values into ticks (degrees of rotation) */
#define HOU_TO_H(hou, min)      (hou)*30 + (min)/2 /* min*6/12 */
#define MIN_TO_M(min, sec)      (min)*6  + (sec)   /* sec*60/60 */
#define SEC_TO_S(sec)           (sec)*6


void *run_clock(void *arg)
{
    window  *win = (window *) arg;
    point    center, radius;
    unsigned hou, min, h, m, s, last_h, last_m, last_s, step;
    double   ex, ey;
    struct tm time;

    assert(win);
    assert(win->bstore);

    /* Push cancellation handler */
    pthread_cleanup_push( (void (*)(void *)) handle_thread_cancel, (void *) win );

    center.x = win->width/2;
    center.y = win->height/2;

    /* Keeping/breaking aspect ratio */
    if ( win->keep_ar ) {
        if ( center.x < center.y ) {
            radius.x = radius.y = (win->width-1)/2 - 3;
            ex       = ey       = ((double) radius.x) / CLOCK_RADIUS;
        } else {
            radius.x = radius.y = (win->height-1)/2 - 3;
            ex       = ey       = ((double) radius.y) / CLOCK_RADIUS;
        }
    } else {
        radius.x = (win->width-1)/2 - 3;
        radius.y = (win->height-1)/2 - 3;
        ex       = ((double) radius.x)/CLOCK_RADIUS;
        ey       = ((double) radius.y)/CLOCK_RADIUS;
    }

    draw_face(win->bstore, win->width, center, radius);
    draw_border(win->bstore, win->width, win->height);

    /* Break up startup_time */
    localtime_r( &win->startup_time, &time );

    /* Set startup times and ticks */
    hou    = time.tm_hour % 12; /* don't need 24 hours */
    min    = time.tm_min;
    s      = SEC_TO_S(time.tm_sec);
    last_h = last_m = last_s = 0;
    step   = 6 / win->resolution;

    while ( 1 ) {
        hou %= 12;
        for ( ; min < 60; min++ ) {
            for ( ; s < 360; s += step ) {
                /* There is no cancellation points between this
                 * pthread_mutex_lock and pthread_mutex_unlock below, so no need
                 * to push unlocking cancellation handler */
                pthread_mutex_lock( &win->mutex );
                /* Erase all hands */
                draw_hand(win->bstore, win->width,
                          hhand, HHAND_LEN, COLOR_BG, -last_h,
                          center.x, center.y, ex, ey);
                draw_hand(win->bstore, win->width,
                          mhand, MHAND_LEN, COLOR_BG, -last_m,
                          center.x, center.y, ex, ey);
                draw_hand(win->bstore, win->width,
                          shand, SHAND_LEN, COLOR_BG, -last_s,
                          center.x, center.y, ex, ey);
                /* Calculate new minutes and hours ticks */
                m = MIN_TO_M(min, s/60);
                h = HOU_TO_H(hou, min);
                /* Draw all hands */
                draw_hand(win->bstore, win->width,
                          hhand, HHAND_LEN, COLOR_HOURS_HAND, -h,
                          center.x, center.y, ex, ey);
                draw_hand(win->bstore, win->width,
                          mhand, MHAND_LEN, COLOR_MINUTES_HAND, -m,
                          center.x, center.y, ex, ey);
                draw_hand(win->bstore, win->width,
                          shand, SHAND_LEN, COLOR_SECONDS_HAND, -s,
                          center.x, center.y, ex, ey);
                pthread_mutex_unlock( &win->mutex );
                /* Remember last ticks */
                last_h = h;
                last_m = m;
                last_s = s;
                /* Take a nap, if necessary */
                pthread_testcancel();
                if ( win->delay.tv_sec || win->delay.tv_nsec ) {
                    nanosleep(&win->delay, NULL);
                    pthread_testcancel();
                }
            }
            s = 0; /* zero out seconds ticks */
        }
        min = MIN_TO_M(0, 0); /* zero out minutes ticks */
        hou++;
    }


    /* We'll never actually reach here because thread never exits.  The code
     * below is here just to match pthread_cleanup_push() above */
    pthread_cleanup_pop(1);

    return NULL;
}



static void draw_face(unsigned char *bstore, int width,
                      point center, point radius)
{
    int x, y, i;

    for ( i = 1; i <= 60; i++ ) {
        x = radius.x * cos(i*6/180.*M_PI - M_PI_2) + center.x;
        y = radius.y * sin(i*6/180.*M_PI - M_PI_2) + center.y;
        if ( i % 5 )
            *(bstore + y*width + x) = COLOR_FACE_DOT_SMALL;
        else {
            *(bstore + y*width + x)     = COLOR_FACE_DOT_BIG;
            *(bstore + y*width + x + 1) = COLOR_FACE_DOT_BIG;
            *(bstore + y*width + x - 1) = COLOR_FACE_DOT_BIG;
            *(bstore + (y+1)*width + x) = COLOR_FACE_DOT_BIG;
            *(bstore + (y-1)*width + x) = COLOR_FACE_DOT_BIG;
        }
    }

    return;
}



static void draw_border(unsigned char *bstore, int width, int height)
{
    int i, j;

    /* Top border */
    for ( i = 0; i < width; i++ )
        *(bstore + i) = COLOR_BORDER_TL;
    /* Left border */
    i = width*height;
    for ( j = 0; j < i; j += width )
        *(bstore + j) = COLOR_BORDER_TL;
    /* Right border */
    if ( width > 0 ) {
        i = width*height;
        for ( j = width - 1; j < i; j+= width )
            *(bstore + j) = COLOR_BORDER_BR;
    }
    /* Bottom border */
    if ( height > 0 ) {
        j = width*height;
        for ( i = width * (height - 1); i < j; i++ )
            *(bstore + i) = COLOR_BORDER_BR;
    }

    return;
}



/*
 * We need to scale the hand in directions of coordinate axes with ratios
 * (ex, ey), rotate it (pi + r) degrees, and then move it by vector (a, b).
 * The entire transformation can be expressed as a 3x3 matrix resulting from
 * multiplication of three 3x3 matrices R(pi + r), T(a, b) and E(ex,ey) for
 * rotation, moving and scaling:
 *
 *  phi = pi + r,
 *
 *         ( cos(phi) sin(phi) 0 )          ( 1 0 a )            ( ex 0  0 )
 *  R(phi)=(-sin(phi) cos(phi) 0 ), T(a, b)=( 0 1 b ), E(ex, ey)=( 0  ey 0 )
 *         (     0        0    1 )          ( 0 0 1 )            ( 0  0  1 )
 *
 * So the new coordinates are:
 *
 *  ( X )   ( 1  0  a )   (  cos(phi)  sin(phi)    0 )   ( ex 0  0 )   ( x )
 *  ( Y ) = ( 0  1  b ) x ( -sin(phi)  cos(phi)    0 ) x ( 0  ey 0 ) x ( y ) =
 *  ( 1 )   ( 0  0  1 )   (      0         0       1 )   ( 0  0  1 )   ( 1 )
 *
 *          (  ex * ( x*cos(phi) + y*sin(phi) )  +  a  )
 *        = (  ey * (-x*sin(phi) + y*cos(phi) )  +  b  )
 *          (                    1                     )
 */
#define TRANSFORM_X          (ex * ( new.x * cos_phi + new.y * sin_phi )  +  a)
#define TRANSFORM_Y          (ey * (-new.x * sin_phi + new.y * cos_phi )  +  b)

static void draw_hand(unsigned char *bstore, int width,
                      const point *hand, int len, char color,
                      int angle, int a, int b, double ex, double ey)
{
    int i;
    point last, new, curr;
    double phi, cos_phi, sin_phi;

    phi = M_PI + angle / 180. * M_PI;
    cos_phi = cos(phi);
    sin_phi = sin(phi);

    new.x  = (hand+len-1)->x;
    new.y  = (hand+len-1)->y;
    last.x = TRANSFORM_X;
    last.y = TRANSFORM_Y;

    for ( i = 0; i < len; i++ ) {
        new.x = (hand+i)->x;
        new.y = (hand+i)->y;
        curr.x = TRANSFORM_X;
        curr.y = TRANSFORM_Y;
        line(bstore, width, color, last.x, last.y, curr.x, curr.y);
        last.x = curr.x;
        last.y = curr.y;
    }

    return;
}



/*
 * Local variables:
 * compile-command:  "make -k"
 * End:
 */
