#include "threads.h"


int not_done_yet = 1;

static void draw_win(unsigned char *scr_store, window *win);
static void redraw_scr_store(unsigned char *bstore0, unsigned char *bstore1);


/* When done, returns 0 (all done) or 1 (memory allocation failed) */
int display_server(void)
{
    unsigned char *bstore0 = NULL, *bstore1 = NULL, *bstore;
    window *win;

    /* Allocate two backing stores for the entire screen */
    bstore0 = (unsigned char *) malloc( scr_width*scr_height );
    if ( !bstore0 )
        return 1;
    bstore1 = (unsigned char *) malloc( scr_width*scr_height );
    if ( !bstore1 ) {
        free( (void *) bstore0 );
        return 1;
    }
    /* Blank them out */
    memset( (void *) bstore0, 0, scr_width*scr_height );
    memset( (void *) bstore1, 0, scr_width*scr_height );

    while ( not_done_yet ) { /* fumbled with in mouse.c or keyboard.c */
        /* We draw all windows in the second backing store, from the last one to
         * the first */
        pthread_mutex_lock(&wlist_mutex);
        win = wlist_tail;
        while ( win ) {
            draw_win(bstore1, win);
            win = win->prev;
        }
        pthread_mutex_unlock(&wlist_mutex);
        /* Draw mouse pointer and process mouse events */
        check_mouse();
        redraw_mouse_pointer(bstore1);
        /* Sync the screen with the second backing store */
        redraw_scr_store(bstore0, bstore1);
        /* Wipe out the first backing store */
        memset( (void *) bstore0, 0, scr_width*scr_height );
        /* Swap the backing stores */
        bstore  = bstore0;
        bstore0 = bstore1;
        bstore1 = bstore;
    }

    free( (void *) bstore0 );
    free( (void *) bstore1 );

    return 0;
}


/* Returns top-most window which occupies point (x, y), or NULL if no window
 * occupies the point.  To make sure the window you get still exists, lock
 * wlist_mutex before calling this function, and unlock it only after you're
 * done working with the window */
window *win_top_xy(int x, int y)
{
    window *win = wlist_head;
    while ( win ) {
        if ( x >= win->x && x < win->x + win->width &&
             y >= win->y && y < win->y + win->height    )
            break;
        win = win->next;
    }

    return win;
}

/* Makes the window top-most (moves it to the head of the window list).  To make
 * sure the window you get still exists, lock wlist_mutex before calling this
 * function, and unlock it only after you're done working with the window */
void win_bring_up(window *win)
{
    if ( !win )
        return;

    if ( win != wlist_head ) {     /* this is not list's head */
        assert(win->prev);
        win->prev->next = win->next;
        if ( win != wlist_tail ) { /* this is not list's tail */
            assert(win->next);
            win->next->prev = win->prev;
        } else {                   /* this is list's tail */
            assert(!win->next);
            wlist_tail = win->prev;
        }
        wlist_head->prev = win;
        win->next = wlist_head;
        win->prev = NULL;
        wlist_head = win;
    }

    return;
}

/* Removes the window from the windows list.  To make sure the window we are
 * unlinking still exists and/or no one uses it while we are unlinking it, lock
 * wlist_mutex before calling this function, and unlock it after this function
 * returns */
void win_unlink(window *win)
{
    int head_or_tail = 0;

    if ( win == wlist_head ) { /* list's head (might also be its tail) */
        assert(!win->prev);
        win->next->prev = NULL;
        wlist_head = win->next;
        head_or_tail = 1;
    }
    if ( win == wlist_tail ) { /* list's tail (might also be its head) */
        assert(!win->next);
        win->prev->next = NULL;
        wlist_tail = win->prev;
        head_or_tail = 1;
    }

/*     if ( win != wlist_head && win != wlist_tail ) { */
    if ( !head_or_tail ) {
        assert(win->prev);
        assert(win->next);
        win->prev->next = win->next;
        win->next->prev = win->prev;
    }

    return;
}


/* To make sure the window we are drawing still exists, lock wlist_mutex before
 * you get pointer to the window and pass it to this function, and unlock it
 * after this function returns */
static void draw_win(unsigned char *scr_store, window *win)
{
    int i, j, width, height;
    int scan_start_offset_scr, scan_start_offset_win;
    int start_x_win, start_y_win, start_x_scr, start_y_scr;
    unsigned char *ptr_scr, *ptr_win;

    assert(win);
    assert(scr_store);

    pthread_mutex_lock( &win->mutex );

    /* Make sure we do not draw past end / before start of scanlines and past
     * end / before start of backing store of the screen.  Here we assume that
     * window is not wider or taller than the screen, and that window is not
     * completely off the screen */
    if ( win->x < 0 ) {
        /* (Part of) window behind left edge of the screen */
        width = win->width + win->x;
        start_x_win = -win->x;
        start_x_scr = 0;
    } else if ( win->x + win->width > scr_width ) {
        /* (Part of) window behind right edge of the screen */
        width = scr_width - win->x;
        start_x_win = 0;
        start_x_scr = win->x;
    } else {
        /* Window fits in the screen (horizontally) */
        width = win->width;
        start_x_win = 0;
        start_x_scr = win->x;
    }
    scan_start_offset_win = win->width - width;
    scan_start_offset_scr = scr_width - width;

    if ( win->y < 0 ) {
        /* (Part of) window above top edge of the screen */
        height = win->height + win->y;
        start_y_win = -win->y;
        start_y_scr = 0;
    } else if ( win->y + win->height > scr_height ) {
        /* (Part of) window below bottom edge of the screen */
        height = scr_height - win->y;
        start_y_win = 0;
        start_y_scr = win->y;
    } else {
        /* Window fits in the screen (vertically) */
        height = win->height;
        start_y_win = 0;
        start_y_scr = win->y;
    }

    /* Initial offsets in window's and screen's backing stores */
    ptr_win = win->bstore + start_y_win * win->width + start_x_win;
    ptr_scr = scr_store + start_y_scr * scr_width + start_x_scr;
    for ( i = 0; i < height; i++ ) {
        for ( j = 0; j < width; j++ )
            *ptr_scr++ = *ptr_win++;
        /* Go to start of next scanline in window's backing store */
        ptr_win += scan_start_offset_win;
        /* Go to start of next scanline in screen's backing store */
        ptr_scr += scan_start_offset_scr;
    }

    pthread_mutex_unlock( &win->mutex );

    return;
}



/* Redraws the screen from image stored in bstore0 to image stored in bstore1,
 * also wiping out bstore0 */
static void redraw_scr_store(unsigned char *bstore0, unsigned char *bstore1)
{
    int x, y;

    for ( y = 0; y < scr_height; y++ )
        for ( x = 0; x < scr_width; x++, bstore0++, bstore1++ )
            if ( *bstore0 != *bstore1 ) {
                vga_setcolor( *bstore1 );
                vga_drawpixel( x, y );
            }

    return;
}



/*
 * Local variables:
 * compile-command:  "make -k"
 * End:
 */
