#include "threads.h"
#include <getopt.h>
#include <ctype.h>


static int startup_threads_count = DEFAULT_STARTUP_THREADS_COUNT;

int     scr_width   = 0;
int     scr_height  = 0;
int     color_depth = 0;
char   *progname    = NULL;
int     resolutions[] = {1, 2, 3, 6};
int     resolutions_count = 0;

/* Windows list and its access mutex */
pthread_mutex_t wlist_mutex;
window *wlist_head  = NULL;
window *wlist_tail  = NULL;

static const char *errMemory   = "%s: error--cannot allocate memory\n";
static const char *errKeyboard = "%s: error--could not initialize keyboard\n";
static const char *errThreadCreation = "%s: error--cannot create a thread\n";
static const char *errMode = "%s: error--mode `%d' not available\n";

static int parse_options(int argc, char **argv);



int main(int argc, char **argv)
{
    int i, j;

    resolutions_count = sizeof(resolutions) / sizeof(resolutions[0]);

    /* Process command line arguments */
    progname = *argv;
    i = parse_options(argc, argv);
    if ( i == 0 )
        return 1;
    else if ( i == 1 )
        return 0;

    /* Seed pseudo-random number generator */
    srand(time(NULL));

    /* Initialize SVGALib */
    vga_init();
    vga_setmousesupport(1);

    i = vga_getdefaultmode();
    if ( i == -1 )
        i = G640x480x256; /* G640x480x16 G1024x768x256 */
    if ( !vga_hasmode(i) ) {
        fprintf( stderr, errMode, progname, i );
        return -1;
    }
    vga_setmode(i);

    scr_width   = vga_getxdim();
    scr_height  = vga_getydim();
    color_depth = vga_getcolors();

    pthread_mutex_init(&wlist_mutex, NULL);
    /* Fire up startup_threads_count threads with random parameters */
    for ( i = 0; i < startup_threads_count; i++ ) {
        j = thread_up(-1, -1, -1, -1, time(NULL), -1, -1, -1);
        if ( j == 1 ) {
            vga_setmode(TEXT);
            fprintf( stderr, errMemory, progname );
            return 1;
        } else if ( j == 2 ) {
            vga_setmode(TEXT);
            fprintf( stderr, errThreadCreation, progname );
            return 1;
        }
    }

    mouse_setwrap(MOUSE_NOWRAP); /* disable wrapping */
    /* We need raw keyboard access */
    if ( keyboard_init() ) {
        vga_setmode(TEXT);
        fprintf( stderr, errKeyboard, progname );
        return 1;
    }

    /* Start drawing windows on the screen */
    if ( display_server() ) {
        keyboard_close();
        vga_setmode(TEXT);
        fprintf( stderr, errMemory, progname );
        return 1;
    }

    /* Close down */
    keyboard_close();
    vga_setmode(TEXT);

    return 0;
}



static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";

static char *errNumThreads    = "%s: error--number of threads must be a positive integer\n";
static char *errInitDelta     = "%s: error--initial delta must be a positive integer\n";
static char *errMaxDelta      = "%s: error--maximum delta must be a positive integer\n";
static char *errDeltaMaxDelta = "%s: error--initial delta cannot be greater than maximum delta\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-n num] [-d delta_init] [-m delta_max]\n\
\n\
-h            display this help message\n\
-n num        number of threads to create on startup\n\
-d delta_init initial delta when emulating mouse with keyboard (default:%4d)\n\
-m delta_max  maximum delta when emulating mouse with keyboard (default:%4d)\n\
\n\
Note:   - to quit, use the keys described below;\n\
        - left-click and drag a window to move it around on the screen;\n\
        - right-click on a window to destroy it;\n\
        - right-click on empty area to create a new window there;\n
        - if you don't have a mouse, use the following keys to emulate mouse\n\
          movement and buttons:\n\
             h, j, k, l           - vi bindings for cursor movement;\n\
             b, n, p, f           - emacs binding for cursor movement;\n\
             keypad keys          - do what you think they do;\n\
             cursor movement keys - do what you think they do;\n\
             <Shift>              - emulates left mouse button;\n\
             <Control>            - emulates right mouse button;\n\
             x, q, <Escape>       - quit the application.
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
static int parse_options(int argc, char **argv)
{
    char c, *ptr;
    int i;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":n:d:m:h")) != -1 )
        switch ( c ) {
        case 'n':
            i = (int) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i <= 0 ) {
                fprintf(stderr, errNumThreads, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            startup_threads_count = i;
            break;
        case 'd':
            i = (int) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i <= 0 ) {
                fprintf(stderr, errInitDelta, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            delta_init = i;
            break;
        case 'm':
            i = (int) strtol(optarg, &ptr, 0);
            if ( ptr == optarg || i <= 0 ) {
                fprintf(stderr, errMaxDelta, progname);
                fprintf(stderr, usehelp, progname);
                return 0;
            }
            delta_max = i;
            break;
        case 'h':
            printf(usage, progname, DEFAULT_DELTA_INIT, DEFAULT_DELTA_MAX);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'n':
                msg = "number of processes";
                break;
            case 'd':
                msg = "initial delta";
                break;
            case 'm':
                msg = "maximum delta";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }


    if ( delta_init > delta_max ) {
        fprintf(stderr, errDeltaMaxDelta, progname);
        fprintf(stderr, usehelp, progname);
        return 0;
    }

    return -1;
}



/*
 * Local variables:
 * compile-command:  "make -k"
 * End:
 */
