#include "threads.h"


#define GET_RAND(min, max)      ( rand() / (RAND_MAX/((max)-(min))) + (min) )


/* Creates a new window thread and places it at the head of windows list.
 * Returns:
 *     0    success;
 *     1    insufficient memory;
 *     2    thread creation error
 */
int thread_up(int x, int y, int width, int height,
              time_t startup_time, int res_idx, long int delay, int keep_ar)
{
    window *win;

    win = (window *) malloc( sizeof(window) );
    if ( !win )
        return 1;

    /* When not getting random dimension, clip it to the valid range */
    if ( width < 0 )
        win->width = GET_RAND(WIN_PLACEMENT_WIDTH_MIN, WIN_PLACEMENT_WIDTH_MAX+1);
    else
        if ( width < WIN_PLACEMENT_WIDTH_MIN )
            win->width = WIN_PLACEMENT_WIDTH_MIN;
        else if ( width > WIN_PLACEMENT_WIDTH_MAX )
            win->width = WIN_PLACEMENT_WIDTH_MAX;
        else
            win->width = width;

    if ( height < 0 )
        win->height = GET_RAND(WIN_PLACEMENT_HEIGHT_MIN, WIN_PLACEMENT_HEIGHT_MAX+1);
    else
        if ( height < WIN_PLACEMENT_HEIGHT_MIN )
            win->height = WIN_PLACEMENT_HEIGHT_MIN;
        else if ( height > WIN_PLACEMENT_HEIGHT_MAX )
            win->height = WIN_PLACEMENT_HEIGHT_MAX;
        else
            win->height = height;

    /* When not getting random coordinate, clip it to the valid rectangle */
    if ( x < 0 )
        win->x = GET_RAND(WIN_PLACEMENT_LEFT_PAD - win->width + 1,
                          scr_width - WIN_PLACEMENT_RIGHT_PAD);
    else
        if ( x <= WIN_PLACEMENT_LEFT_PAD - win->width )
            win->x = WIN_PLACEMENT_LEFT_PAD - win->width + 1;
        else if ( x >= scr_width - WIN_PLACEMENT_RIGHT_PAD )
            win->x = scr_width - WIN_PLACEMENT_RIGHT_PAD - 1;
        else
            win->x = x;

    /* When not getting random coordinate, clip it to the valid rectangle */
    if ( y < 0 )
        win->y = GET_RAND(WIN_PLACEMENT_TOP_PAD - win->height + 1,
                          scr_height - WIN_PLACEMENT_BOTTOM_PAD);
    else
        if ( y <= WIN_PLACEMENT_TOP_PAD - win->height )
            win->y = WIN_PLACEMENT_TOP_PAD - win->height + 1;
        else if ( y >= scr_height - WIN_PLACEMENT_BOTTOM_PAD )
            win->y = scr_height - WIN_PLACEMENT_BOTTOM_PAD - 1;
        else
            win->y = y;

    win->startup_time = startup_time;

    if ( res_idx < 0 || res_idx >= resolutions_count )
        win->resolution = *(resolutions + GET_RAND(0, resolutions_count));
    else
        win->resolution = *(resolutions + res_idx);

    win->delay.tv_sec = 0;
    if ( delay < 0 || delay >= 1000000000 )
        win->delay.tv_nsec = GET_RAND(0, 1000000000);
    else
        win->delay.tv_nsec = delay;

    if ( keep_ar < 0 || keep_ar > 1 )
        win->keep_ar = GET_RAND(0, 2);
    else
        win->keep_ar = keep_ar;

    /* Create window's backing store and blank it out.  This is in case it
     * display server picks it up before the new thread has a chance to draw the
     * clock in it--this way the worst we get is a black window, instead of the
     * white noise */
    win->bstore = (char *) malloc( win->width * win->height );
    if ( !win->bstore ) {
        free(win);
        return 1;
    }
    memset( (void *) win->bstore, 0, win->width * win->height );

    pthread_mutex_init(&win->mutex, NULL);

    /* Add new window to the end of windows list */
    pthread_mutex_lock(&wlist_mutex);
    win->next        = wlist_head;
    win->prev        = NULL;
    if ( wlist_head )
        wlist_head->prev = win;
    wlist_head       = win;
    if ( !wlist_tail )
        wlist_tail = win;
    pthread_mutex_unlock(&wlist_mutex);

    if ( pthread_create(&win->tid, NULL, run_clock, (void *) win) ) {
        /* Thread not created--clean up and return */
        free( win->bstore );
        free( win );
        return 2;
    }

    return 0;
}


/* Thread's cancellation handler */
void *handle_thread_cancel(void *arg)
{
    window *win = (window *) arg;

    if ( win ) {
        pthread_mutex_lock(&wlist_mutex);
        win_unlink(win);
        pthread_mutex_unlock(&wlist_mutex);

        pthread_mutex_destroy(&win->mutex);
        free( (void *) win->bstore );
        free( (void *) win );
    }

    return NULL;
}


/* Do not lock wlist_mutex before calling this function--handle_thread_cancel
 * will lock it */
int thread_down(window *win)
{
    if ( !win )
        return 0;

    return pthread_cancel(win->tid);
}



/*
 * Local variables:
 * compile-command:  "make -k"
 * End:
 */
