#include "wired.h"
#include <stdio.h>
#include <stdlib.h>
#include <gtk/gtk.h>
#include <getopt.h>
#include <ctype.h>
#include <assert.h>


const char *errMemory  = "%s: error allocating memory\n";
const char *errParsing = "%s: error parsing input\n";
const char *errNumVert = "%s: number of vertices must be a positive integer\n";
const char *errOrphanEdge = "%s: orphan edge (%d, %d) detected in the input\n";

void *xmalloc(int size)
{
    void *ptr = malloc(size);
    if ( !ptr ) {
        fprintf(stderr, errMemory, progname);
        gtk_exit(1);
    }
    return ptr;
}


static void parsing_error(void)
{
    fprintf(stderr, errParsing, progname);
    gtk_exit(1);
}

/* Searches triangles list to find (at most) two triangles which own the given
 * edge.  Stops as soon as the second triangle is found */
static void edge_find_owning_triangles(Edge *edge, Triangle *triangles)
{
    assert(edge);
    edge->t2 = edge->t1 = NULL;
    while ( triangles ) {
        if ( ( triangles->p1 == edge->p1 ||
               triangles->p2 == edge->p1 ||
               triangles->p3 == edge->p1    ) &&
             ( triangles->p1 == edge->p2 ||
               triangles->p2 == edge->p2 ||
               triangles->p3 == edge->p2    )    ) {
            if ( edge->t1 ) {
                edge->t2 = triangles;
                return;
            } else
                edge->t1 = triangles;
        }
        triangles = triangles->next;
    }

    return;
}

Wired *read_data(FILE *stream)
{
    Wired *obj = (Wired *) xmalloc(sizeof(Wired));
    int i, p1, p2, p3, color;

    /* Read view point and number of vertices */
    if ( 9 != fscanf(stream, "%le %le %le %le %d %d %d %d %d",
                      &obj->viewpoint.x, &obj->viewpoint.y, &obj->viewpoint.z,
                      &obj->screen,
                      &obj->incr_x, &obj->incr_y, &obj->incr_z, &obj->incr_s,
                      &obj->num_vertices) )
        parsing_error();
    if ( obj->num_vertices <= 0 ) {
        fprintf(stderr, errNumVert, progname);
        gtk_exit(1);
    }
    /* Allocate and read in array of num_vertices points */
    obj->vertices = (Point *) xmalloc(sizeof(Point) * obj->num_vertices);
    for ( i = 0; i < obj->num_vertices; i++ )
        if ( 3 != fscanf(stream, "%le %le %le",
                         &(obj->vertices + i)->x,
                         &(obj->vertices + i)->y,
                         &(obj->vertices + i)->z) )
            parsing_error();
    /* Read in the list of triangulated faces (it is separated by a negative
     * number from the list of edges) */
    obj->triangles = NULL;
    while ( 1 == (i = fscanf(stream, "%d", &p1)) && p1 >= 0 ) {
        Triangle *ptr;
        if ( 2 != fscanf(stream, "%d %d", &p2, &p3) )
            parsing_error();
        ptr = (Triangle *) xmalloc(sizeof(Triangle));
        ptr->p1 = p1, ptr->p2 = p2, ptr->p3 = p3;
        if ( color < 0 ) color = 0; else if ( color > 15 ) color = 15;
        ptr->next      = obj->triangles;
        obj->triangles = ptr;
    }
    if ( i != 1 ) /* never read the separating negative number */
        parsing_error();
    /* Read in the list of edges */
    obj->edges = NULL;
    while ( 2 == fscanf(stream, "%d %d", &p1, &p2) ) {
        Edge *ptr = (Edge *) xmalloc(sizeof(Edge));
        ptr->p1 = p1,   ptr->p2 = p2;
        ptr->next  = obj->edges;
        obj->edges = ptr;
        edge_find_owning_triangles(obj->edges, obj->triangles);
        if ( !obj->edges->t1 ) { /* no triangle owns this edge */
            fprintf( stderr, errOrphanEdge, progname, obj->edges->p1, obj->edges->p2 );
            gtk_exit(1);
        }
    }

    return obj;
}



extern FILE *stream_in;
extern int   projection;
static char *fname_in         = NULL;
static char *errUnknownOption = "%s: error--unknown option `-%c'\n";
static char *errUnknownOptionChar = "%s: error--unknown option character `\\x%x'\n";
static char *errArgumentReq   = "%s: error--option `-%c' requires an argument (%s)\n";
static char *errOpenFile      = "%s: cannot open file `%s' for reading\n";

static char *usehelp = "Try `%s -h' for help\n";
static char *usage   = "\n\
Usage: %s [-h] [-i <fname>] [-l]\n\
\n\
Displays perspective or parallel projection of wire model of a 3D object,\n\
allowing to interactively adjust the view.\n\
\n\
-h            display this help message\n\
-i <fname>    file to read data from, `-' for stdin (default: stdin)\n\
-l            do parallel projection (default: do perspective projection)\n\
\n\
Note: parameters of the 3D object are the following sequence of ASCII integers:\n\
      - three world coordinates of viewpoint;\n\
      - position of the screen on z axis of view coordinate system;\n\
      - three movement deltas (along world coordinate sysem's axes);\n\
      - one movement delta for the screen (along z axis of view CS);\n\
      - number of vertices the object has;\n\
      - triplets of coordinates for the vertices;\n\
      - triplets of vertex numbers (triangles) forming triangulated faces of\n\
        the object.  The triangles must be described counterclockwise when\n\
        observing a face from the outside;\n\
      - separator (a negative number);\n\
      - list of object's edges as couples of vertex numbers.\n\
\n\
      Center of world coordinate system is assumed to be the object's center.\n\
\n\
      Use the following keys (keypad also works) for moving viewpoint around\n\
      the object / zooming:\n\
      - End   / PgDn        +/- viewpoint along world CS's x-axis;\n\
      - Right / Left        +/- viewpoint along world CS's y-axis;\n\
      - Up    / Down        +/- viewpoint along world CS's z-axis;\n\
      - Home  / PgUp        +/- screen along view CS's z-axis (zooming);\n\
      - q     / Escape      quit the program\n\
\n";

/* Return codes:
 * 0    - error, exit(1);
 * 1    - ok, exit(0);
 * else - ok, go on.
 */
int parse_options(int argc, char **argv)
{
    char c;

    opterr = 0; /* prevent getopt from printing error messages */
    while ( (c = getopt(argc, argv, ":i:lh")) != -1 )
        switch ( c ) {
        case 'i':
            if ( *optarg == '-' && *(optarg+1) =='\0' )
                fname_in = NULL;
            else
                fname_in = optarg;
            break;
        case 'l':
            projection = 0;
            break;
        case 'h':
            printf(usage, progname);
            return 1;
            break;
        case '?': /* unknown option */
            if ( isprint(optopt) ) {
                fprintf(stderr, errUnknownOption, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            else {
                fprintf(stderr, errUnknownOptionChar, progname, optopt);
                fprintf(stderr, usehelp, progname);
            }
            return 0;
            break;
        case ':': { /* missing option argument */
            char *msg = NULL;

            switch ( optopt ) {
            case 'i':
                msg = "input file name";
                break;
            default:
                assert(0); /* should never get here */
                abort();
                break;
            }
            fprintf(stderr, errArgumentReq, progname, optopt, msg);
            fprintf(stderr, usehelp, progname);
            return 0;
            break;
        }
        default:
            assert(0); /* should never happen */
            abort();
            break;
        }

    if ( !fname_in )
        stream_in = stdin;
    else {
        stream_in = fopen(fname_in, "r");
        if ( !stream_in ) {
            fprintf(stderr, errOpenFile, progname, fname_in);
            exit(1);
        }
    }


    return -1;
}



/*
 * Local variables:
 * compile-command:  "make -k && ./wired < kybos.data"
 * End:
 */
