#include "wired.h"
#include <stdlib.h>
#include <math.h>
#include <assert.h>


Point *recalc_into_viewcs( Wired *obj )
{
    Point *viewvertices;
    int i;
    double cos_phi, sin_phi, cos_theta, sin_theta;
    double proj_xy, ro;

    assert(obj);
    viewvertices = (Point *) xmalloc(sizeof(Point) * obj->num_vertices);

    /* Recalculate Cartesian coordinates of the viewpoint into spherical
       coordinates */
    ro = sqrt( obj->viewpoint.x * obj->viewpoint.x +
               obj->viewpoint.y * obj->viewpoint.y +
               obj->viewpoint.z * obj->viewpoint.z );
    proj_xy = sqrt( obj->viewpoint.x * obj->viewpoint.x +
                    obj->viewpoint.y * obj->viewpoint.y );
    cos_phi   = obj->viewpoint.z / ro;
    sin_phi   = proj_xy / ro;
    cos_theta = obj->viewpoint.x / proj_xy;
    sin_theta = obj->viewpoint.y / proj_xy;
    /* Recalculate world coordinates of the object into the view coordinates */
    for ( i = 0; i < obj->num_vertices; i++ ) {
        double x = (obj->vertices + i)->x;
        double y = (obj->vertices + i)->y;
        double z = (obj->vertices + i)->z;
        (viewvertices + i)->x =
            -x * sin_theta
            + y * cos_theta;
        (viewvertices + i)->y =
            - x * cos_phi * cos_theta
            - y * cos_phi * sin_theta
            + z * sin_phi;
        (viewvertices + i)->z =
            - x * sin_phi * cos_theta
            - y * sin_phi * sin_theta
            - z * cos_phi + ro;
    }

    return viewvertices;
}


#define IS_VISIBLE_PERSPECTIVE(P1, P2, P3)                  \
                    ( (P1)->x * (P2)->y * (P3)->z +         \
                      (P2)->x * (P3)->y * (P1)->z +         \
                      (P3)->x * (P1)->y * (P2)->z -         \
                      (P3)->x * (P2)->y * (P1)->z -         \
                      (P2)->x * (P1)->y * (P3)->z -         \
                      (P1)->x * (P3)->y * (P2)->z   > 0 )

#define IS_VISIBLE_PARALLEL(P1, P2, P3)                                     \
                    ( ( (P2)->x - (P1)->x ) * ( (P3)->y - (P1)->y ) -       \
                      ( (P2)->y - (P1)->y ) * ( (P3)->x - (P1)->x )   > 0 )

void draw_object( GdkDrawable *drawable, Wired *obj )
{
    Point    *viewvertices;
    Triangle *curr;
    Edge     *edge;
    gint      width, height;

    assert(obj);
    viewvertices = recalc_into_viewcs( obj );

    gdk_window_get_size( drawable, &width, &height );
    width  /= 2;
    height /= 2;

    curr = obj->triangles;
    if ( obj->projection ) {    /* do the perspective projection */
        /* Determine visibility of each triangle */
        while ( curr ) {
            Point *p1 = viewvertices + curr->p1;
            Point *p2 = viewvertices + curr->p2;
            Point *p3 = viewvertices + curr->p3;
            if ( IS_VISIBLE_PERSPECTIVE(p1, p2, p3) )
                curr->visible = 1;
            else
                curr->visible = 0;
            curr = curr->next;
        }
        /* Draw edges */
        edge = obj->edges;
        while ( edge ) {
            Point *p1 = viewvertices + edge->p1;
            Point *p2 = viewvertices + edge->p2;
            GdkGC *gc;

            assert(edge->t1);
            if ( edge->t1->visible || (edge->t2 && edge->t2->visible) )
                /* One of the owning triangles is visible--draw solid */
                gc = obj->gc_solid;
            else
                /* None of the owning triangles is visible--draw dashed */
                gc = obj->gc_dashed;
            /* Flip the picture vertically (hence `-' for the y coordinate) */
            gdk_draw_line( drawable, gc,
                           obj->screen * p1->x / p1->z + width,
                           -obj->screen * p1->y / p1->z + height,
                           obj->screen * p2->x / p2->z + width,
                           -obj->screen * p2->y / p2->z + height );
            edge = edge->next;
        }
    } else {                    /* do the parallel projection */
        /* Determine visibility of each triangle */
        while ( curr ) {
            Point *p1 = viewvertices + curr->p1;
            Point *p2 = viewvertices + curr->p2;
            Point *p3 = viewvertices + curr->p3;
            if ( IS_VISIBLE_PARALLEL(p1, p2, p3) )
                curr->visible = 1;
            else
                curr->visible = 0;
            curr = curr->next;
        }
        /* Draw edges */
        edge = obj->edges;
        while ( edge ) {
            Point *p1 = viewvertices + edge->p1;
            Point *p2 = viewvertices + edge->p2;
            GdkGC *gc;

            assert(edge->t1);
            if ( edge->t1->visible || (edge->t2 && edge->t2->visible) )
                /* One of the owning triangles is visible--draw solid */
                gc = obj->gc_solid;
            else
                /* None of the owning triangles is visible--draw dashed */
                gc = obj->gc_dashed;
            /* Flip the picture vertically (hence `-' for the y coordinate) */
            gdk_draw_line( drawable, gc,
                           p1->x + width,
                           -p1->y + height,
                           p2->x + width,
                           -p2->y + height );
            edge = edge->next;
        }
    }

    free(viewvertices);

    return;
}



/*
 * Local variables:
 * compile-command:  "make -k && ./wired < kybos.data"
 * End:
 */
